/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/fld/access.c#1 $
*/
#include <string.h>

#include <avs/dll_in.h>
#include <avs/arr.h>
#include <avs/om.h>
#include <avs/om_type.h>
#include <avs/om_att.h>
#include <avs/fld.h>

typedef  unsigned char byte;

static float IXFORM[16] = {1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0};
/*-------------------------------*/
/*      GENERAL FUNCTIONS        */
/*-------------------------------*/

OMobj_id FLDadd_elem(field, temp_name, elem_name)
    OMobj_id  field;
    char *temp_name, *elem_name;
{
	OMobj_id elem;
	
	elem = OMfind_subobj (OMtempl_obj, OMstr_to_name(temp_name),
				 OM_OBJ_RD);
	if (OMis_null_obj(elem)) 
		return(OMnull_obj);
	elem = OMcopy_obj(elem, OMstr_to_name(elem_name), field, 
			   OMlocal_proc_id, OM_COPY_CLOSE);
	/*
	 * The system currently does not figure out that this is not
	 * supposed to be saved so we must do it by hand.
	 */
	OMset_obj_atts(elem,OM_atts_nosave);
	return(elem);
}

int FLDadd_int(field, name)
    OMobj_id  field;
    char *name;
{
	OMobj_id elem;

	elem = OMcreate_obj(OMstr_to_name(name), OMstr_to_type("int"), field, OMnull_obj,
			     OMlocal_proc_id);
	if (OMis_null_obj(elem)) 
		return(-1);
	return(1);
}

int FLDset_int(field, name, val)
    OMobj_id  field;
    char    *name;
    int     val;
{
	return
		( OMset_name_int_val (field, OMstr_to_name(name), val) );
}

int FLDget_int(field, name, val)
    OMobj_id  field;
    char    *name;
    int     *val;
{
	return
		( OMget_name_int_val (field, OMstr_to_name(name), val) );
}

int FLDadd_real(field, name)
    OMobj_id  field;
    char *name;
{
	OMobj_id elem;

	elem = OMcreate_obj(OMstr_to_name(name), OMstr_to_type("double"), field, OMnull_obj,
			     OMlocal_proc_id);
	if (OMis_null_obj(elem)) 
		return(-1);
	return(1);
}

int FLDset_real(field, name, val)
    OMobj_id  field;
    char    *name;
    double     val;
{
	return
		(OMset_name_real_val (field, OMstr_to_name(name), (double)val));
}

int FLDget_real(field, name, val)
    OMobj_id  field;
    char    *name;
    double     *val;
{
	double dval;
	int    stat;
	stat = OMget_name_real_val (field, OMstr_to_name(name), &dval);
	if (stat < 1)
		return (stat);
	*val = (float)dval;
}

int FLDadd_int_array(field, name, size)
    OMobj_id  field;
    char *name;
    int  size;
{
	int   stat;
	OMobj_id elem, delem, dims_elem;

	elem = OMcreate_obj(OMstr_to_name(name), OMstr_to_type("int"), field, OMnull_obj,
			     OMlocal_proc_id);
	if (OMis_null_obj(elem)) 
		return(-1);
	delem = OMcreate_obj(OM_NULL_NAME, OM_type_raw_int, OMnull_obj,
				OMnull_obj, OMlocal_proc_id);
	if (OMis_null_obj(delem)) 
		return(-1);
	OMset_int_val(delem, size);
	dims_elem = OMcreate_obj(OM_name_dims, OM_type_val_list, 
				  elem, OMnull_obj, OMlocal_proc_id);
	if (OMis_null_obj(dims_elem)) 
		return(-1);
	stat = OMadd_subobj(dims_elem, delem);
	if (stat<0)
		return(-1);
	return(1);
}

int FLDset_array_int (field, name, size, array, mode)
    OMobj_id  field;
    int size, mode;
    int *array;
    char *name;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)array, size, mode);
	return(stat);
}

int  FLDget_array_int (field, name, array, size, mode)
    OMobj_id  field;
    int     **array;
    int     *size, mode;
    char     *name;
{
	OMobj_id  elem_id;
	int        type, stat;

	*array = (int *)NULL;
	*size = 0;
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)array, size, mode);
	return (stat);
}



int FLDadd_float_array(field, name, size)
    OMobj_id  field;
    char *name;
    int  size;
{
	int   stat;
	OMobj_id elem, delem, dims_elem;

	elem = OMcreate_obj(OMstr_to_name(name), OMstr_to_type("float"), field, OMnull_obj,
			     OMlocal_proc_id);
	if (OMis_null_obj(elem)) 
		return(-1);
	delem = OMcreate_obj(OM_NULL_NAME, OM_type_raw_int, OMnull_obj,
				OMnull_obj, OMlocal_proc_id);
	if (OMis_null_obj(delem)) 
		return(-1);
	OMset_int_val(delem, size);
	dims_elem = OMcreate_obj(OM_name_dims, OM_type_val_list, elem, 
				  OMnull_obj, OMlocal_proc_id);
	if (OMis_null_obj(dims_elem)) 
		return(-1);
	stat = OMadd_subobj(dims_elem, delem);
	return(stat);
}


int FLDset_array_float (field, name, size, array, mode)
    OMobj_id  field;
    int size, mode;
    float *array;
    char *name;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)array, size, mode);
	return(stat);
}


int  FLDget_array_float (field, name, array, size, mode)
    OMobj_id  field;
    float     **array;
    int     *size, mode;
    char     *name;
{
	OMobj_id  elem_id;
	int        type, stat;

	*array = (float *)0;
	*size = 0;
	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)array, size, mode);
	return (stat);
}

int FLDget_array_size(e, size)
    OMobj_id  e;
    int *size;
{
	return(OMget_array_size(e, size));
}

int  FLDequal(e1, e2)
    OMobj_id  e1, e2;
{
	OMobj_id  v1, v2;
	int  stat;

	if ((stat=OMget_obj_val(e1, &v1)) != 1 ||
	    (stat=OMget_obj_val(e2, &v2)) != 1 )
		return (stat);
	return(OMequal_objs(v1, v2));
}

int FLDget_parent_val(e, parent)
    OMobj_id  e, *parent;
{
	OMobj_id  e1;
	int  stat;

	if ((stat=OMget_obj_val(e, &e1)) != 1)
		return (stat);
	return(OMget_obj_parent(e1, parent));
}

int FLDget_subelem(field, name, elem)
    OMobj_id  field, *elem;
    char *name;
{

	*elem = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(*elem)) 
		return(0);
	return(1);
}

int FLDget_subelem_seq(field, temp_name)
    OMobj_id  field;
    char    *temp_name;
{
	OMobj_id  elem;

	elem = OMfind_subobj (OMtempl_obj, OMstr_to_name(temp_name),
				   OM_OBJ_RD);
	if (OMis_null_obj(elem)) 
		return(0);
	return(OMget_obj_seq (field, elem, OM_SEQ_VAL));
}

OMobj_id  FLDcreate(parent_id, field_desc, name)
     OMobj_id parent_id;
    char *field_desc, *name;
{
	OMobj_id  parnt, elem, elem_mrg;
	char  *desc, *token;

	if (OMis_null_obj(parent_id))
		parnt = OMinst_obj;
	else
		parnt = parent_id;
	desc = (char *)malloc((strlen(field_desc)+1)*sizeof(char));
	strcpy(desc, field_desc);
	token = strtok(desc, " ");
	if (token == NULL)
		return(OMnull_obj);
	elem = OMfind_subobj (OMtempl_obj, OMstr_to_name(token),
				   OM_OBJ_RD);
	if (OMis_null_obj(elem)) 
		return(OMnull_obj);
	elem = OMcopy_obj(elem, OMstr_to_name(name), parnt, OMlocal_proc_id, 
			   OM_COPY_CLOSE);
	token = strtok(NULL, " ");
	while (token) {
		elem_mrg = OMfind_subobj (OMtempl_obj, OMstr_to_name(token),
					   OM_OBJ_RD);
		if (OMis_null_obj(elem_mrg)) 
			return(OMnull_obj);
		OMmerge_obj(elem_mrg, elem, 0);
		token = strtok(NULL, " ");
	}
	free(desc);	      
	return(elem);
}

int FLDdestroy(field)
    OMobj_id  field;
{
	return(OMdestroy_obj(field,0));
}

int FLDdelete(field, name)
    OMobj_id  field;
    char *name;
{
	OMobj_id elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name(name), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMdestroy_obj(elem_id,0));
}

int  FLDcopy_elem_ref(ref, base)
    OMobj_id  base, ref;
{
	int  stat;

	stat =  OMset_obj_ref_mode(ref, OM_OBJ_REF);
	if (stat < 1)
		return(stat);
	stat = OMset_obj_ref(ref, base, 0);
	if (stat < 1)
		return(stat);
	return(1);
}

int  FLDcopy_subelem_ref(ref, base)
    OMobj_id  base, ref;
{
	int  stat;

	/*
	 * We want the NO_MATCH mode here both for performance and also
	 * because we can't rely upon the connections matching properly.
	 * If there is old data in the field, we may end up setting
	 * values before nvals is set... the connections will not match
	 * in this case.
	 */
	stat = OMset_obj_val(ref, base, OM_OBJ_VAL_REF_MODE | 
					 OM_OBJ_REF_RDONLY | 
					 OM_OBJ_REF_NO_MATCH);
	return(stat);
}

int FLDget_function(elem, name, func)
    OMobj_id  elem;
    char  *name;
    OMpfi *func;
{
	OMobj_id  func_id;
	int  stat;

	func_id = OMfind_subobj (elem, OMstr_to_name(name), OM_OBJ_RD);
	if (OMis_null_obj(func_id)) 
		return(0);
	stat = OMget_func_val(func_id, func);
	return(stat);
}

int FLDget_methods(elem, n, names, func)
    OMobj_id  elem;
    int n;
    char  *names[];
    OMpfi *func;
{
	OMobj_id  func_id, elem_id;
	int  i, stat;

	elem_id = OMfind_subobj (elem, OMstr_to_name("methods"), OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) {
		for (i=0; i<n; i++)
			func[i] = (OMpfi)0;
		return(1);	
	}
	for (i=0; i<n; i++) {
		func_id = OMfind_subobj (elem_id, OMstr_to_name(names[i]), OM_OBJ_RD);
		if (!OMis_null_obj(func_id)) {
			stat = OMget_func_val(func_id, &(func[i]));
			if (stat != 1)
				func[i] = (OMpfi)0;
		}
		else 
			func[i] = (OMpfi)0;
	}
	return(1);
}

/*-----------------------------------------------*/
/*           GENERAL FIELD                       */
/*-----------------------------------------------*/
int FLDget_grid_type(field, grid_type)
    OMobj_id  field;
    int     *grid_type;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("grid_type"), grid_type) );
}
int FLDset_grid_type (field, grid_type)
	OMobj_id  field;
	int grid_type;
{
	return
		(OMset_name_int_val (field, OMstr_to_name("grid_type"), grid_type));
}

int FLDget_ndim(field, ndim)
    OMobj_id  field;
    int     *ndim;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("ndim"), ndim) );
}
int FLDset_ndim (field, ndim)
	OMobj_id  field;
	int ndim;
{
	return
		(OMset_name_int_val (field, OMstr_to_name("ndim"), ndim));
}

int FLDget_nnodes(field, nnodes)
    OMobj_id  field;
    int     *nnodes;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("nnodes"), nnodes) );
}

int FLDset_nnodes (field, nnodes)
	OMobj_id  field;
	int nnodes;
{
	int  stat;

	stat = OMset_name_int_val (field, OMstr_to_name("nnodes"), nnodes);
	return(stat);
}

int  FLDget_dims (field, dims, size)
    OMobj_id  field;
    int     **dims;
    int     *size;
{
	OMobj_id  elem_id;
	int        type, stat;
	int     mode;

	mode = OM_GET_ARRAY_RD;
	*dims = (int *)NULL;
	*size = 0;
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (field, OMstr_to_name("dims"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)dims, size, mode);
	return (stat);
}

int FLDset_dims (field, dims)
	OMobj_id  field;
	int *dims;
{
	int   stat, type, ndim;
	OMobj_id   elem_id;

	OMget_name_int_val (field, OMstr_to_name("ndim"), &ndim);
	
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (field, OMstr_to_name("dims"), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)dims, ndim, OM_SET_ARRAY_COPY);
	return(stat);
}

/*--------------------------------------------------*/
/*         COORDINATES                              */
/*--------------------------------------------------*/

int FLDget_xform(field, xform)
    OMobj_id  field;
    float      *xform;
{

/*************************************************
	int i;
	float tmp_mat[16], tmp_x[3];
	OMobj_id  elem_id;
	
	memcpy(xform, IXFORM, 16*sizeof(float));
	elem_id = OMfind_subobj (field, OMstr_to_name("xform"), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) {
		return(0);
	}
	GDxform_get_matrix(elem_id, tmp_mat);
	GDxform_get_xlate(elem_id, tmp_x);
	for (i=0; i<3;i++)
		memcpy(xform+4*i, tmp_mat+4*i, 3*sizeof(float));
	memcpy(xform+12, tmp_x, 3*sizeof(float));
************************************************/
	OMobj_id  elem_id;

	memcpy(xform, IXFORM, 16*sizeof(float));
	elem_id = OMfind_subobj (field, OMstr_to_name("xform"), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) {
		return(1);
	}
	GDxform_get_comp_matrix(elem_id, xform);
	
	return(1);
}

int FLDset_xform(field, xform)
    OMobj_id  field;
    float      *xform;
{
	OMobj_id  elem_id;
	int  i;
	float tmp_mat[16], tmp_x[3];

	elem_id = OMfind_subobj (field, OMstr_to_name("xform"), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) {
		return(0);
	}
	memcpy(tmp_mat, IXFORM, 16*sizeof(float));
	for (i=0; i<3;i++)
		memcpy(tmp_mat+4*i, xform+4*i, 3*sizeof(float));
	memcpy(tmp_x, xform+12, 3*sizeof(float));
	GDxform_set_matrix(elem_id, tmp_mat);
	GDxform_set_xlate(elem_id, tmp_x);
	return(1);
}

int FLDget_rspace(field, rspace)
    OMobj_id  field;
    int	       *rspace;
{
	OMobj_id  elem_id;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("xform"), OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) {
		return(0);
	}
	return (OMget_name_int_val (elem_id, OMstr_to_name("rspace"), rspace));
}

int FLDset_rspace(field, rspace)
    OMobj_id  field;
    int	       rspace;
{
	OMobj_id  elem_id;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("xform"), OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) {
		return(0);
	}
	return (OMset_name_int_val (elem_id, OMstr_to_name("rspace"), rspace));
}

int FLDget_npoints(field, npoints)
    OMobj_id  field;
    int     *npoints;
{
	return( OMget_name_int_val (field, OMstr_to_name("npoints"), npoints) );
}

int FLDset_npoints(field, npoints)
    OMobj_id  field;
    int     npoints;
{
	return( OMset_name_int_val (field, OMstr_to_name("npoints"), npoints) );
}

int FLDget_nspace(field, nspace)
    OMobj_id  field;
    int     *nspace;
{
	return( OMget_name_int_val (field, OMstr_to_name("nspace"), nspace) );
}

int FLDset_nspace (field, nspace)
	OMobj_id  field;
	int nspace;
{
	return (OMset_name_int_val (field, OMstr_to_name("nspace"), nspace));
}

int FLDget_coord(field, coord_array, size, mode)
    OMobj_id  field;
    int     *size;
    float    **coord_array;
    int     mode;
{
	OMobj_id  elem_id;
	int    type, stat, nnodes, nspace;

	type = OM_TYPE_FLOAT;
	*coord_array = (float *)NULL;
	*size = 0;
	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"), 
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if ((stat = OMget_name_int_val (field, OMstr_to_name("nnodes"), &nnodes)) != 1)
		return(stat);

	if ((stat = OMget_name_int_val (field, OMstr_to_name("nspace"), &nspace)) != 1)
		return(stat);
	elem_id = OMfind_subobj (elem_id, OMstr_to_name("values"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if ((stat=OMget_array_sz(elem_id, &type, (char **)coord_array, 
				 size, mode)) != 1)
		return(stat);
	if (nnodes*nspace != *size)
		return(0);
	return(1);
}

int FLDget_sub_coord(field, ndim, dims, min_rng, max_rng, coord_array)
    OMobj_id  field;
    int       ndim, *dims, *min_rng, *max_rng;
    float    *coord_array;
{
	OMobj_id  elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"), 
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id = OMfind_subobj (elem_id, OMstr_to_name("values"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return (OMget_sub_farray(elem_id, ndim, dims, min_rng, max_rng,
				 (void *)coord_array));
}

int FLDset_coord(field, coord_array, size, mode)
    OMobj_id  field;
    int size, mode;
    float  *coord_array;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id = OMfind_subobj (elem_id, OMstr_to_name("values"), 
				    OM_OBJ_RW);
	stat = OMset_array(elem_id, type, (char *)coord_array, size, mode);
	return(stat);
}

int FLDreset_point_extent(field)
    OMobj_id  field;
{
	int   stat;
	OMobj_id   elem_id, elem_id_min, elem_id_max;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id_min = OMfind_subobj (elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMparse_obj_ref(elem_id_min, "cache(min_array(points, 0, 0))");
	if (stat != 1)
		return(stat);
	stat = OMparse_obj_ref(elem_id_max, "cache(max_array(points, 0, 0))");
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDreset_coord_extent(field)
    OMobj_id  field;
{
	int   stat;
	OMobj_id   elem_id, elem_id_min, elem_id_max;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id_min = OMfind_subobj (elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMparse_obj_ref(elem_id_min, "cache(min_array(values, 0, 0))");
	if (stat != 1)
		return(stat);
	stat = OMparse_obj_ref(elem_id_max, "cache(max_array(values, 0, 0))");
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDset_coord_extent(field, min_extent, max_extent, nspace)
    OMobj_id  field;
    float   *min_extent, *max_extent;
    int nspace;
{
	int   stat, type;
	OMobj_id   elem_id, elem_id_min, elem_id_max;

	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id_min = OMfind_subobj (elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMset_array(elem_id_min, type, (char *)min_extent, 
			   nspace, OM_SET_ARRAY_COPY);
	if (stat != 1)
		return(stat);
	stat = OMset_array(elem_id_max, type, (char *)max_extent, 
			   nspace, OM_SET_ARRAY_COPY);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDget_coord_extent(field, min_extent, max_extent)
    OMobj_id  field;
    float   *min_extent, *max_extent;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max;
	int    type, stat, size, mode, nspace, npoints, stat1, stat2;
	float  *tmin, *tmax;
	
	if (FLDget_nspace(field, &nspace) != 1)
		return(-1);
	type = OM_TYPE_FLOAT;
	size = 0;
	mode = OM_GET_ARRAY_RD;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	
	elem_id_min = OMfind_subobj (elem_id, OMstr_to_name("min_vec"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (elem_id, OMstr_to_name("max_vec"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	mode = OM_GET_ARRAY_RD;
	tmin = (float *)NULL;
	tmax = (float *)NULL;
	size = 0;
	stat1 = OMget_array_sz(elem_id_min, &type, (char **)&tmin, &size, mode);
	size = 0;
	stat2 = OMget_array_sz(elem_id_max, &type, (char **)&tmax, &size, mode);
	/*
	 * Make sure to free one of these if the other happens not to be
	 * valid
	 */
	if (stat1 == 1) {
	   memcpy(min_extent, tmin, nspace*sizeof(float));
	   if (tmin != NULL) ARRfree(tmin);
	}
	if (stat2 == 1) {
	   memcpy(max_extent, tmax, nspace*sizeof(float));
	   if (tmax != NULL) ARRfree(tmax);
	}
	if (stat1 != 1 || stat2 != 1)
		return(-1);
	return (1);
}

int  FLDget_points (field, points, size, mode)
    OMobj_id  field;
    float   **points;
    int     *size;
    int     mode;
{
	OMobj_id  elem_id;
	int        type, stat;

 	/* XXX Hack to work around dgroup bug with npoints and rect flds. */
	(void) OMfind_subobj(field, OMstr_to_name("npoints"), OM_OBJ_RW);

	*points = (float *)NULL;
	*size = 0;
	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name("points"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);

	stat = OMget_array_sz(elem_id, &type, (char **)points, size, mode);
	return (stat);
}



int FLDset_points(field, points, size, mode)
    OMobj_id  field;
    int size, mode;
    float  *points;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (field, OMstr_to_name("points"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(FLDset_coord(field, points, size, mode));
	stat = OMset_array(elem_id, type, (char *)points, size, mode);
	return(stat);
}

int FLDget_coord_units(field, units, size)
    OMobj_id  field;
    char    *units;
    int     size;
{
	OMobj_id    elem_id;

	/* Check for programmer error: */
	if (units == NULL || size == 0)
	   return 0;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);

	elem_id = OMfind_subobj (elem_id, OMstr_to_name("units"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &units, size));
}


int FLDset_coord_units (field, units)
	OMobj_id  field;
	char *units;
{
	int   stat;
	OMobj_id   elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("coordinates"), 
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	elem_id = OMfind_subobj (elem_id, OMstr_to_name("units"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val(elem_id, units);
	return(stat);
}


/*-------------------------------------------------------------*/
/*                  NODE DATA                                  */
/*-------------------------------------------------------------*/

int FLDget_node_data_ncomp(field, nnode_comp)
    OMobj_id  field;
    int     *nnode_comp;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("nnode_data"), nnode_comp) );
}

int FLDset_node_data_ncomp (field, nnode_comp)
	OMobj_id  field;
	int nnode_comp;
{
	return(OMset_name_int_val (field, OMstr_to_name("nnode_data"), nnode_comp));
}

int FLDset_node_data_comp (field, comp, veclen, label, units)
	OMobj_id  field;
	int   comp, veclen;
	char *label, *units;
{
	OMobj_id  elem_id, sub_elem_id;
	int   stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("labels"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val(elem_id, label);
	if (stat < 1)
		return (0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("units"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val(elem_id, units);
	if (stat < 1)
		return (0);
	return(1);
}


int FLDget_node_data_veclen(field, comp, veclen)
    OMobj_id  field;
    int     comp;
    int     *veclen;
{
	OMobj_id  elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RD) <= 0)
		return(0);
	return
		( OMget_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen) );
}

int FLDset_node_data_veclen(field, comp, veclen)
    OMobj_id  field;
    int     comp;
    int     veclen;
{
	OMobj_id  elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	return(OMset_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen));
}

int FLDget_node_data_type(field, comp, type)
    OMobj_id  field;
    int     comp;
    int     *type;
{
	OMobj_id  elem_id, sub_elem_id;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_data_type(elem_id, type));
}

int FLDset_node_data_type(field, comp, type)
    OMobj_id  field;
    int     comp;
    int     type;
{
	int stat;

	OMobj_id  elem_id, sub_elem_id;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	/* set values */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set null_value */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("null_value"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set min max */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set min_vec max_vec */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDget_node_data(field, comp, type, node_data, size, mode)
    OMobj_id  field;
    int     comp;
    int     *type, *size;
    char    **node_data;
    int     mode;
{
	OMobj_id  elem_id, sub_elem_id;
	int   stat, nnodes, veclen;

	*node_data = (char *)NULL;
	*size = 0;
	if (mode != OM_GET_ARRAY_WR)
		*type = OM_TYPE_UNSET;
	if ((stat = OMget_name_int_val (field, OMstr_to_name("nnodes"), &nnodes)) != 1)
		return(stat);

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if ((stat=OMget_name_int_val (sub_elem_id, OMstr_to_name("veclen"), 
				      &veclen)) != 1)
		return(stat);
	if ((stat = OMget_array_sz(elem_id, type, (char **)node_data, 
				   size, mode)) != 1)
		return(stat);
	if (veclen*nnodes != *size)
		return(0);
	return(1);
}

int FLDget_typed_node_data(field, comp, type, node_data, size, mode)
    OMobj_id  field;
    int     comp;
    int     *type, *size;
    char    **node_data;
    int     mode;
{
	OMobj_id  elem_id, sub_elem_id;
	int   stat, nnodes, veclen, dtype;

	*node_data = (char *)NULL;
	*size = 0;
	if ((stat = OMget_name_int_val (field, OMstr_to_name("nnodes"), &nnodes)) != 1)
		return(stat);

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if ((stat=OMget_name_int_val (sub_elem_id, OMstr_to_name("veclen"), 
				      &veclen)) != 1)
		return(stat);
	if ((stat = OMget_array_sz(elem_id, type, (char **)node_data, 
				   size, mode)) != 1)
		return(stat);
	if (veclen*nnodes != *size)
		return(0);
	return(1);
}

int FLDget_sub_node_data(field, comp, type, ndim, dims, min_rng, max_rng, node_data)
    OMobj_id  field;
    int     comp;
    int       ndim, *dims, *min_rng, *max_rng;
    int     *type;
    char    *node_data;
{
	int  stat;
	OMobj_id  elem_id, sub_elem_id;

	if (*type == OM_TYPE_UNSET) {
		stat = FLDget_node_data_type(field, comp, type);
		if (stat < 0)
			return(0);
	}
	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	switch (*type) {
	      case OM_TYPE_DOUBLE:
		return (OMget_sub_rarray(elem_id, ndim, dims, min_rng, max_rng,
					 (double *)node_data));
	      case OM_TYPE_FLOAT:
		return (OMget_sub_farray(elem_id, ndim, dims, min_rng, max_rng,
					 (float *)node_data));
	      case OM_TYPE_INT:
		return (OMget_sub_iarray(elem_id, ndim, dims, min_rng, max_rng,
					 (int *)node_data));
	      case OM_TYPE_SHORT:
		return (OMget_sub_sarray(elem_id, ndim, dims, min_rng, max_rng,
					 (short *)node_data));
	      case OM_TYPE_BYTE:
		return (OMget_sub_barray(elem_id, ndim, dims, min_rng, max_rng,
					 (byte *)node_data));
	      case OM_TYPE_CHAR:
		return (OMget_sub_barray(elem_id, ndim, dims, min_rng, max_rng,
					 (unsigned char *)node_data));
	      default:
		return(0);
	}
}


int FLDset_node_data(field, comp, node_data, type, size, mode)
    OMobj_id  field;
    int type, size, mode;
    int comp;
    char  *node_data;
{
	int   stat;
	OMobj_id   elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if ((stat = OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW))!= 1)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"), 
				    OM_OBJ_RW);
	stat = OMset_array(elem_id, type, (char *)node_data, size, mode);
	return(stat);
}

int FLDget_node_data_minmax_vec(field, comp, min, max, type, mode)
    OMobj_id  field;
    int     comp;
    char   **min, **max;
    int     *type;
    int     mode;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int    stat1, stat2, size;

	if (mode != OM_GET_ARRAY_WR)
		*type = OM_TYPE_UNSET;
	*min = (char *)NULL;
	*max = (char *)NULL;
	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	size = 0;
	stat1 = OMget_array_sz(elem_id_min, type, (char **)min, &size, mode);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	size = 0;
	stat2 = OMget_array_sz(elem_id_max, type, (char **)max, &size, mode);
	if (stat1 != 1 || stat2 != 1)
		return(-1);
	return(1);
}

int FLDcopy_node_minmax_vec(in_field, out_field, in_comp, out_comp)
    OMobj_id  in_field, out_field;
    int in_comp, out_comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	OMobj_id  elem_id1, elem_id_min1, elem_id_max1, sub_elem_id1;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, in_comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	elem_id1 = OMfind_subobj (out_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id1)) 
		return(0);
	if (OMget_array_val (elem_id1, out_comp, &sub_elem_id1, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id1)) 
		return(0);

	elem_id_min1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min1)) 
		return(0);
	elem_id_max1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max1)) 
		return(0);
 	/*
	 * The RDONLY flag here is necessary to prevent the "write" flag
	 * from propagating past us.  It will also prevent assignments
	 * through this connection which is probably good too...
	 */
	stat = OMset_obj_ref(elem_id_min1, elem_id_min, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	stat = OMset_obj_ref(elem_id_max1, elem_id_max, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDget_node_data_minmax(field, comp, min, max)
    OMobj_id  field;
    int     comp;
    char   *min, *max;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int   stat, stat1, stat2;
	int  type, val_int;
	double val_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	
	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	stat = OMget_data_type(elem_id_min, &type);
	if (stat != 1)
		return(stat);
	switch (type) {
	      case OM_TYPE_CHAR:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((char *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((char *)max) = val_int;
		break;
	      case OM_TYPE_BYTE:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((unsigned char *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((unsigned char *)max) = val_int;
		break;
	      case OM_TYPE_SHORT:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((short *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((short *)max) = val_int;
		break;
	      case OM_TYPE_INT:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((int *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((int *)max) = val_int;
		break;
	      case OM_TYPE_FLOAT:
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    &val_real);
		/*
		 * Cannot dereference an uninitialized floating point value
		 * or a crash may result on some systems.
		 */
		if (stat != 1) 
		   return(stat);
		*((float *)min) = val_real;
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    &val_real);
		if (stat != 1) 
		   return(stat);
		*((float *)max) = val_real;
		break;
	      case OM_TYPE_DOUBLE:
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    &val_real);
		if (stat != 1) 
		   return(stat);
		*((double *)min) = val_real;
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    &val_real);
		if (stat != 1) 
		   return(stat);
		*((double *)max) = val_real;
		break;
	      default:
		stat = -1;
		break;
	}
	return(stat);
}

int FLDset_node_data_minmax(field, comp, min, max, dtype)
    OMobj_id  field;
    int     comp, dtype;
    char   *min, *max;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max,  sub_elem_id;
	int   stat, stat1, stat2;
	int  type, val_int;
	double val_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	
	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	stat = OMget_data_type(elem_id_min, &type);
	if (stat != 1)
		return(stat);
	if (type == OM_TYPE_UNSET || type != dtype) {
		type =dtype;
	}
	stat = OMset_data_type(elem_id_min, type);
	if (stat != 1)
		return(stat);
	stat = OMset_data_type(elem_id_max, type);
	if (stat != 1)
		return(stat);
	
	switch (type) {
	      case OM_TYPE_CHAR:
		val_int = *((char *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((char *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_BYTE:
		val_int = *((unsigned char *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((unsigned char *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_SHORT:
		val_int = *((short *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((short *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_INT:
		val_int = *((int *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((int *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_FLOAT:
		val_real = *((float *)min);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    val_real);
		val_real = *((float *)max);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    val_real);
		break;
	      case OM_TYPE_DOUBLE:
		val_real = *((double *)min);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    val_real);
		val_real = *((double *)max);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    val_real);
		break;
	      default:
		stat = -1;
		break;
	}
	return(stat);
}

int FLDcopy_node_minmax(in_field, out_field, in_comp, out_comp)
    OMobj_id  in_field, out_field;
    int in_comp, out_comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	OMobj_id  elem_id1, elem_id_min1, elem_id_max1, sub_elem_id1;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, in_comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	elem_id1 = OMfind_subobj (out_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id1)) 
		return(0);
	if (OMget_array_val (elem_id1, out_comp, &sub_elem_id1, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id1)) 
		return(0);

	elem_id_min1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min1)) 
		return(0);
	elem_id_max1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max1)) 
		return(0);
	stat = OMset_obj_ref(elem_id_min1, elem_id_min, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	stat = OMset_obj_ref(elem_id_max1, elem_id_max, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDreset_node_minmax(in_field, comp)
    OMobj_id  in_field;
    int comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMparse_obj_ref(elem_id_min, 
               "cache(min_array(magnitude(values), null_flag, null_value))");
	if (stat != 1)
		return(stat);
	stat = OMparse_obj_ref(elem_id_max,
		"cache(max_array(magnitude(values), null_flag, null_value))");
	if (stat != 1)
		return(stat);

	return(1);
}

int FLDreset_node_minmax_vec(in_field, comp)
    OMobj_id  in_field;
    int comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMparse_obj_ref(elem_id_min, 
		"cache(min_array(values, null_flag, null_value))");
	if (stat != 1)
		return(stat);
	stat = OMparse_obj_ref(elem_id_max,
		"cache(max_array(values, null_flag, null_value))");
	if (stat != 1)
		return(stat);

	return(1);
}

int FLDget_node_data_label(field, comp, label, size)
    OMobj_id  field;
    int     comp;
    int     size;
    char    *label;
{
	OMobj_id    elem_id, sub_elem_id;

	/* Check for programmer error: */
	if (label == NULL || size == 0)
	   return 0;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("labels"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &label, size));
}

int FLDget_node_data_units(field, comp, units, size)
    OMobj_id  field;
    int     comp;
    int     size;
    char    *units;
{
	OMobj_id    elem_id, sub_elem_id;

	/* Check for programmer error: */
	if (units == NULL || size == 0)
	   return 0;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("units"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &units, size));
}

int FLDget_node_data_id(field, comp, id)
	OMobj_id  field;
	int     comp;
	int     *id;
{
	OMobj_id    elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
		( OMget_name_int_val (sub_elem_id, OMstr_to_name("id"), id) );
}

int FLDset_node_data_id(field, comp, id)
	OMobj_id  field;
	int     comp;
	int     id;
{
	OMobj_id    elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
		( OMset_name_int_val (sub_elem_id, OMstr_to_name("id"), id) );
}

int FLDget_node_null_data(field, comp, null_flag, null_value)
    OMobj_id  field;
    int     comp;
    int     *null_flag;
    char    *null_value;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat, size, type, null_int;
	double   null_real;

	null_int = 0;
	null_real = 0.0;
	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RD) <= 0)
		return(0);
	stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), null_flag);

	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				  OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	
	stat = OMget_data_type(elem_id, &type);
	if (stat != 1)
		return(stat);

	if (*null_flag) {
		switch (type) {
		      case OM_TYPE_CHAR:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((char *)null_value) = null_int;
			break;
		      case OM_TYPE_BYTE:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((unsigned char *)null_value) = null_int;
			break;
		      case OM_TYPE_SHORT:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((short *)null_value) = null_int;
			break;
		      case OM_TYPE_INT:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((int *)null_value) = null_int;
			break;
		      case OM_TYPE_FLOAT:
			stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
						    &null_real);
			if (stat == 1)
			    *((float *)null_value) = null_real;
			break;
		      case OM_TYPE_DOUBLE:
			stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
						    &null_real);
			if (stat == 1)
			   *((double *)null_value) = null_real;
			break;
		      default:
			stat = -1;
			break;
		}
	}
	else {
		stat = 1;
		switch (type) {
		      case OM_TYPE_CHAR:
			*((char *)null_value) = null_int;
			break;
		      case OM_TYPE_BYTE:
			*((unsigned char *)null_value) = null_int;
			break;
		      case OM_TYPE_SHORT:
			*((short *)null_value) = null_int;
			break;
		      case OM_TYPE_INT:
			*((int *)null_value) = null_int;
			break;
		      case OM_TYPE_FLOAT:
			*((float *)null_value) = null_real;
			break;
		      case OM_TYPE_DOUBLE:
			*((double *)null_value) = null_real;
			break;
		      default:
			stat = -1;
			break;
		}
	}
	return(stat);
}

int FLDset_node_null_data(field, comp, null_value, type)
    OMobj_id  field;
    int     comp, type;
    char    *null_value;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat, size, dtype, null_int;
	double   null_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), 1);
	if (stat != 1)
		return(stat);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("null_value"),
				    OM_OBJ_RW);


	stat = OMget_data_type(elem_id, &dtype);
	if (stat != 1)
		return(stat);
	if (dtype == OM_TYPE_UNSET) {
		stat = OMset_data_type(elem_id, type);
		if (stat != 1)
			return(stat);
	}


	switch (type) {
	      case OM_TYPE_CHAR:
		null_int = *((char *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_BYTE:
		null_int = *((unsigned char *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_SHORT:
		null_int = *((short *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_INT:
		null_int = *((int *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_FLOAT:
		null_real = *((float *)(null_value));
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
					     null_real);
		break;
	      case OM_TYPE_DOUBLE:
		null_real = *((double *)(null_value));
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
					    null_real);
		break;
	      default:
		stat = -1;
		break;
	}

	return(stat);
}

int FLDset_node_null_flag(field, comp, null_flag)
    OMobj_id  field;
    int     comp, null_flag;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("node_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
	      (OMset_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), null_flag));
}

/*-------------------------------------------------------------*/
/*                  CELL DATA                                  */
/*-------------------------------------------------------------*/

int FLDget_cell_data_ncomp(field, ncell_comp)
    OMobj_id  field;
    int     *ncell_comp;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("ncell_data"), ncell_comp) );
}

int FLDset_cell_data_ncomp (field, ncell_comp)
	OMobj_id  field;
	int ncell_comp;
{
	return(OMset_name_int_val (field, OMstr_to_name("ncell_data"), ncell_comp));
}

int FLDset_cell_data_comp (field, comp, veclen, label, units)
	OMobj_id  field;
	int   comp, veclen;
	char *label, *units;
{
	OMobj_id  elem_id, sub_elem_id;
	int   stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("labels"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val(elem_id, label);
	if (stat < 1)
		return (0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("units"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val(elem_id, units);
	if (stat < 1)
		return (0);
	return(1);
}


int FLDset_cell_data_veclen(field, comp, veclen)
    OMobj_id  field;
    int     comp;
    int     veclen;
{
	OMobj_id  elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	return(OMset_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen));
}

int FLDget_cell_data_veclen(field, comp, veclen)
    OMobj_id  field;
    int     comp;
    int     *veclen;
{
	OMobj_id  elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RD) <= 0)
		return(0);
	return
		( OMget_name_int_val (sub_elem_id, OMstr_to_name("veclen"), veclen) );
}

int FLDget_cell_data_type(field, comp, type)
    OMobj_id  field;
    int     comp;
    int     *type;
{
	OMobj_id  elem_id, sub_elem_id;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_data_type(elem_id, type));
}

int FLDset_cell_data_type(field, comp, type)
    OMobj_id  field;
    int     comp;
    int     type;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat;
	
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	/* set values */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set null_value */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("null_value"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set min max */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);

	/* set min_vec max_vec */
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_data_type(elem_id, type);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDget_cell_data(field, comp, type, cell_data, size, mode)
    OMobj_id  field;
    int     comp;
    int     *type, *size;
    char    **cell_data;
    int     mode;
{
	OMobj_id  elem_id, sub_elem_id;

	*cell_data = (char *)NULL;
	*size = 0;
	if (mode != OM_GET_ARRAY_WR)
		*type = OM_TYPE_UNSET;
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return (OMget_array_sz(elem_id, type, (char **)cell_data, size, mode));
}

int FLDget_typed_cell_data(field, comp, type, cell_data, size, mode)
    OMobj_id  field;
    int     comp;
    int     *type, *size;
    char    **cell_data;
    int     mode;
{
	OMobj_id  elem_id, sub_elem_id;

	*cell_data = (char *)NULL;
	*size = 0;
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return (OMget_array_sz(elem_id, type, (char **)cell_data, size, mode));
}


int FLDset_cell_data(field, comp, cell_data, type, size, mode)
    OMobj_id  field;
    int type, size, mode;
    int comp;
    char  *cell_data;
{
	int   stat;
	OMobj_id   elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"), 
				    OM_OBJ_RW);
	stat = OMset_array(elem_id, type, (char *)cell_data, size, mode);
	return(stat);
}

int FLDget_cell_data_minmax_vec(field, comp, min, max, type, mode)
    OMobj_id  field;
    int     comp;
    char   **min, **max;
    int     *type;
    int     mode;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int    stat1, stat2, size;

	if (mode != OM_GET_ARRAY_WR)
		*type = OM_TYPE_UNSET;
	*min = (char *)NULL;
	*max = (char *)NULL;
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	size = 0;
	stat1 = OMget_array_sz(elem_id_min, type, (char **)min, &size, mode);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	size = 0;
	stat2 = OMget_array_sz(elem_id_max, type, (char **)max, &size, mode);
	if (stat1 != 1 || stat2 != 1)
		return(-1);
	return(1);
}

int FLDcopy_cell_minmax_vec(in_field, out_field, in_comp, out_comp)
    OMobj_id  in_field, out_field;
    int in_comp, out_comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	OMobj_id  elem_id1, elem_id_min1, elem_id_max1, sub_elem_id1;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, in_comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	elem_id1 = OMfind_subobj (out_field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id1)) 
		return(0);
	if (OMget_array_val (elem_id1, out_comp, &sub_elem_id1, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id1)) 
		return(0);

	elem_id_min1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("min_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min1)) 
		return(0);
	elem_id_max1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("max_vec"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max1)) 
		return(0);
	stat = OMset_obj_ref(elem_id_min1, elem_id_min, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	stat = OMset_obj_ref(elem_id_max1, elem_id_max, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	return(1);
}

int FLDreset_cell_minmax(cell_set, comp)
    OMobj_id  cell_set;
    int comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int stat;

	elem_id = OMfind_subobj (cell_set, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);
	stat = OMparse_obj_ref(elem_id_min, 
               "cache(min_array(magnitude(values), null_flag, null_value))");
	if (stat != 1)
		return(stat);
	stat = OMparse_obj_ref(elem_id_max,
		"cache(max_array(magnitude(values), null_flag, null_value))");
	if (stat != 1)
		return(stat);

	return(1);
}


int FLDget_cell_data_minmax(field, comp, min, max)
    OMobj_id  field;
    int     comp;
    char   *min, *max;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int   stat, stat1, stat2;
	int  type, val_int;
	double val_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	
	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	stat = OMget_data_type(elem_id_min, &type);
	if (stat != 1)
		return(stat);

	switch (type) {
	      case OM_TYPE_CHAR:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((char *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((char *)max) = val_int;
		break;
	      case OM_TYPE_BYTE:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((unsigned char *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((unsigned char *)max) = val_int;
		break;
	      case OM_TYPE_SHORT:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((short *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((short *)max) = val_int;
		break;
	      case OM_TYPE_INT:
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   &val_int);
		*((int *)min) = val_int;
		stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   &val_int);
		*((int *)max) = val_int;
		break;
	      case OM_TYPE_FLOAT:
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    &val_real);
		if (stat != 1)
		     return(stat);
		*((float *)min) = val_real;
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    &val_real);
		if (stat != 1)
		     return(stat);
		*((float *)max) = val_real;
		break;
	      case OM_TYPE_DOUBLE:
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    &val_real);
		if (stat != 1)
		     return(stat);
		*((double *)min) = val_real;
		stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    &val_real);
		if (stat != 1)
		     return(stat);
		*((double *)max) = val_real;
		break;
	      default:
		stat = -1;
		break;
	}
	return(stat);
}

int FLDget_cell_data_minmax_total(field, comp, min, max)
    OMobj_id  field;
    int comp;
    char   *min, *max;
{
	int stat, nsets, set, type;
	OMobj_id  tmp;
	double rmin, rmax, dmin, dmax;

	rmin=0.0; rmax=0.0; type = 0;

	stat = FLDget_ncell_sets(field, &nsets);
	if (stat != 1)
		return(stat);
	for (set=0; set<nsets; set++) {
		stat = FLDget_cell_set(field, set, &tmp);
		if (stat != 1)
			return(stat);
		stat = FLDget_cell_data_type(tmp, comp, &type);
		if (stat != 1)
			continue;
		stat = FLDget_cell_data_minmax(tmp, comp, min, max);
		if (stat != 1)
			return(stat);
		UTILtype_to_double(&dmax, max, type);
		UTILtype_to_double(&dmin, min, type);
		if (set == 0) {
			rmin = dmin;
			rmax = dmax;
		}
		else {
			if (dmin < rmin)
				rmin = dmin;
			if (dmax > rmax)
				rmax = dmax;
		}
	}
	UTILdouble_to_type(min, rmin, type);
	UTILdouble_to_type(max, rmax, type);
	return(1);
}

int FLDset_cell_data_minmax(field, comp, min, max, dtype)
    OMobj_id  field;
    int     comp, dtype;
    char   *min, *max;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	int   stat, stat1, stat2;
	int  type, val_int;
	double val_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				  OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	
	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				      OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);

	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	stat = OMget_data_type(elem_id_min, &type);
	if (stat != 1)
		return(stat);
	
	if (type == OM_TYPE_UNSET || type != dtype) {
		type =dtype;
	}
	stat = OMset_data_type(elem_id_min, type);
	if (stat != 1)
		return(stat);
	stat = OMset_data_type(elem_id_max, type);
	if (stat != 1)
		return(stat);

	switch (type) {
	      case OM_TYPE_CHAR:
		val_int = *((char *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((char *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_BYTE:
		val_int = *((unsigned char *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((unsigned char *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_SHORT:
		val_int = *((short *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((short *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_INT:
		val_int = *((int *)min);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("min"), 
					   val_int);
		val_int = *((int *)max);
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("max"), 
					   val_int);
		break;
	      case OM_TYPE_FLOAT:
		val_real = *((float *)min);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    val_real);
		val_real = *((float *)max);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    val_real);
		break;
	      case OM_TYPE_DOUBLE:
		val_real = *((double *)min);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("min"), 
					    val_real);
		val_real = *((double *)max);
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("max"), 
					    val_real);
		break;
	      default:
		stat = -1;
		break;
	}
	return(stat);
}

int FLDcopy_cell_minmax(in_field, out_field, in_comp, out_comp)
    OMobj_id  in_field, out_field;
    int in_comp, out_comp;
{
	OMobj_id  elem_id, elem_id_min, elem_id_max, sub_elem_id;
	OMobj_id  elem_id1, elem_id_min1, elem_id_max1, sub_elem_id1;
	int stat;

	elem_id = OMfind_subobj (in_field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, in_comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id_min = OMfind_subobj (sub_elem_id, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min)) 
		return(0);
	elem_id_max = OMfind_subobj (sub_elem_id, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max)) 
		return(0);

	elem_id1 = OMfind_subobj (out_field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id1)) 
		return(0);
	if (OMget_array_val (elem_id1, out_comp, &sub_elem_id1, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id1)) 
		return(0);

	elem_id_min1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("min"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_min1)) 
		return(0);
	elem_id_max1 = OMfind_subobj (sub_elem_id1, OMstr_to_name("max"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id_max1)) 
		return(0);
	stat = OMset_obj_ref(elem_id_min1, elem_id_min, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	stat = OMset_obj_ref(elem_id_max1, elem_id_max, OM_OBJ_REF_RDONLY);
	if (stat != 1)
		return(stat);
	return(1);
}



int FLDget_cell_data_label(field, comp, label, size)
    OMobj_id  field;
    int     comp;
    int     size;
    char    *label;
{
	OMobj_id    elem_id, sub_elem_id;

	/* Check for programmer error: */
	if (label == NULL || size == 0)
	   return 0;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("labels"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &label, size));
}

int FLDget_cell_data_units(field, comp, units, size)
    OMobj_id  field;
    int     comp;
    int     size;
    char    *units;
{
	OMobj_id    elem_id, sub_elem_id;

	/* Check for programmer error: */
	if (units == NULL || size == 0)
	   return 0;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("units"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &units, size));
}

int FLDget_cell_data_id(field, comp, id)
	OMobj_id  field;
	int     comp;
	int     *id;
{
	OMobj_id    elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
		( OMget_name_int_val (sub_elem_id, OMstr_to_name("id"), id) );
}

int FLDset_cell_data_id(field, comp, id)
	OMobj_id  field;
	int     comp;
	int     id;
{
	OMobj_id    elem_id, sub_elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
		( OMset_name_int_val (sub_elem_id, OMstr_to_name("id"), id) );
}

int FLDget_cell_null_data(field, comp, null_flag, null_value)
    OMobj_id  field;
    int     comp;
    int     *null_flag;
    char    *null_value;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat, size, type, null_int;
	double   null_real;

	null_int = 0;
	null_real = 0.0;
	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RD) <= 0)
		return(0);
	stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), null_flag);

	if (stat != 1)
		return(stat);
	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("values"),
				  OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	
	stat = OMget_data_type(elem_id, &type);
	if (stat != 1)
		return(stat);

	if (*null_flag) {
		switch (type) {
		      case OM_TYPE_CHAR:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((char *)null_value) = null_int;
			break;
		      case OM_TYPE_BYTE:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((unsigned char *)null_value) = null_int;
			break;
		      case OM_TYPE_SHORT:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((short *)null_value) = null_int;
			break;
		      case OM_TYPE_INT:
			stat = OMget_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
						   &null_int);
			*((int *)null_value) = null_int;
			break;
		      case OM_TYPE_FLOAT:
			stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
						    &null_real);
			if (stat == 1)
			   *((float *)null_value) = null_real;
			break;
		      case OM_TYPE_DOUBLE:
			stat = OMget_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
						    &null_real);
			if (stat == 1)
			   *((double *)null_value) = null_real;
			break;
		      default:
			stat = -1;
			break;
		}
	}
	else {
		stat = 1;
		switch (type) {
		      case OM_TYPE_CHAR:
			*((char *)null_value) = null_int;
			break;
		      case OM_TYPE_BYTE:
			*((unsigned char *)null_value) = null_int;
			break;
		      case OM_TYPE_SHORT:
			*((short *)null_value) = null_int;
			break;
		      case OM_TYPE_INT:
			*((int *)null_value) = null_int;
			break;
		      case OM_TYPE_FLOAT:
			*((float *)null_value) = null_real;
			break;
		      case OM_TYPE_DOUBLE:
			*((double *)null_value) = null_real;
			break;
		      default:
			stat = -1;
			break;
		}
	}
	return(stat);
}

int FLDset_cell_null_data(field, comp, null_value, type)
    OMobj_id  field;
    int     comp;
    char    *null_value;
    int	    type;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat, size, dtype, null_int;
	double   null_real;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), 1);
	if (stat != 1)
		return(stat);

	elem_id = OMfind_subobj (sub_elem_id, OMstr_to_name("null_value"),
				    OM_OBJ_RW);

	stat = OMget_data_type(elem_id, &dtype);
	if (stat != 1)
		return(stat);
	if (dtype == OM_TYPE_UNSET) {
		stat = OMset_data_type(elem_id, type);
		if (stat != 1)
			return(stat);
	}

	switch (type) {
	      case OM_TYPE_CHAR:
		null_int = *((char *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_BYTE:
		null_int = *((unsigned char *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_SHORT:
		null_int = *((short *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_INT:
		null_int = *((int *)(null_value));
		stat = OMset_name_int_val (sub_elem_id, OMstr_to_name("null_value"), 
					   null_int);
		break;
	      case OM_TYPE_FLOAT:
		null_real = *((float *)(null_value));
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
					     null_real);
		break;
	      case OM_TYPE_DOUBLE:
		null_real = *((double *)(null_value));
		stat = OMset_name_real_val (sub_elem_id, OMstr_to_name("null_value"), 
					    null_real);
		break;
	      default:
		stat = -1;
		break;
	}

	return(stat);
}

int FLDset_cell_null_flag(field, comp, null_flag)
    OMobj_id  field;
    int     comp, null_flag;
{
	OMobj_id  elem_id, sub_elem_id;
	int stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_data"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, comp, &sub_elem_id, OM_OBJ_RW) <= 0)
		return(0);
	if (OMis_null_obj(sub_elem_id)) 
		return(0);
	return
	      (OMset_name_int_val (sub_elem_id, OMstr_to_name("null_flag"), null_flag));
}


/*--------------------------------------------------*/
/*         CELL SET                                 */
/*--------------------------------------------------*/

int FLDget_cell_set(field, i, cell_set)
    OMobj_id  field, *cell_set;
    int     i;
{
	OMobj_id  elem_id, sub_elem_id;
	OMtype_id type;
	OMobj_name name;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_set"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	if (OMget_array_val (elem_id, i, cell_set, OM_OBJ_RW) <= 0)
		return(-1);
	if (OMget_obj_val_user(*cell_set, cell_set) != 1) return(-1);
#if 0
	/* This hack shouldn't be needed anymore; I changed the call above
	   to be OMget_obj_val_user, which never returns value objects (it
	   look upward until it finds a non-value object).  See bug 5752.
	   */
	/*
	 * This is a gross hack for bugs 3393 and 3394.  The problem is
	 * that the OMobj_val call is going through and getting the id
	 * of the merge object which is a value.  We need some mode
	 * on the get_obj_val call that will stop right before the
	 * merge for this case since the merge is not a rooted object.
	 *
	 * The attempt to add a ref to it screws up big time.
	 */
	OMget_obj_type(*cell_set,&type);
	OMget_obj_name(*cell_set,&name);
	if (!strcmp(type->name,"merge") && name == OM_NULL_NAME) 
	    *cell_set = OMret_obj_parent(*cell_set);
#endif
        return(1);	 
}

int FLDget_cell_set_name(field, name, size)
    OMobj_id  field;
    char      *name;
    int        size;
{
	OMobj_id  elem_id;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_name"),
				    OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &name, size));
}

int FLDget_cell_set_user_name(field, name, size)
    OMobj_id  field;
    char      *name;
    int        size;
{
	OMobj_id  elem_id;
	int  stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("name"),
				    OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	return(OMget_str_val (elem_id, &name, size));
}

int FLDset_cell_set_user_name(field, name)
    OMobj_id  field;
    char      *name;
{
	OMobj_id  elem_id;
	int  stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("name"),
				    OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_str_val (elem_id, name);
        return(stat);	 
}

int FLDset_cell_set(cell_set, cell_set_name)
    OMobj_id  cell_set;
    char       *cell_set_name;
{
	OMobj_id  elem, corner_id;
	int       stat, order;


	elem = OMfind_subobj (OMtempl_obj, OMstr_to_name(cell_set_name),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem)) 
		return(0);

	/**************************************************/
	/* workaround of problem with breaking connection */
	/*        cell_corner_nnodes => cell_nnodes       */
	/* for second order cells                         */
	/**************************************************/
	corner_id = OMfind_subobj (cell_set, OMstr_to_name("cell_corner_nnodes"),
				   OM_OBJ_RW);
	if (!OMis_null_obj(corner_id)) {
		if (OMget_name_int_val(elem, OMstr_to_name("cell_order"), &order) != OM_STAT_SUCCESS)
			order = 0;
		if (order > 1) {
			stat = OMset_obj_ref(corner_id, OMnull_obj, 0);
			if (stat != OM_STAT_SUCCESS)
				return(0);
		}
	}
	stat = OMmerge_obj(elem, cell_set, 0);
	return(stat);
}

int FLDadd_cell_set(field, cell_set_name)
    OMobj_id  field;
    char       *cell_set_name;
{
	OMobj_id  elem_id, elem, val_id, corner_id;
	int size, stat, order;


	elem_id = OMfind_subobj (field, OMstr_to_name("cell_set"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);

	if ((stat = OMget_array_size(elem_id, &size)) != 1) {
		size = 0;
	}
	if ((stat = OMset_array_size(elem_id,size+1)) != 1) {
		return(0);
	}

	elem = OMfind_subobj (OMtempl_obj, OMstr_to_name(cell_set_name),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem)) 
		return(0);

	if ((stat = OMget_array_val (elem_id, size, &val_id, OM_OBJ_RW)) != 1) {
		return(0);
	}
	/**************************************************/
	/* workaround of problem with breaking connection */
	/*        cell_corner_nnodes => cell_nnodes       */
	/* for second order cells                         */
	/**************************************************/
	corner_id = OMfind_subobj (val_id, OMstr_to_name("cell_corner_nnodes"),
				   OM_OBJ_RW);
	if (!OMis_null_obj(corner_id)) {
		if (OMget_name_int_val(elem, OMstr_to_name("cell_order"), &order) != OM_STAT_SUCCESS)
			order = 0;
		if (order > 1) {
			stat = OMset_obj_ref(corner_id, OMnull_obj, 0);
			if (stat != OM_STAT_SUCCESS)
				return(0);
		}
	}
	stat = OMmerge_obj(elem, val_id, 0);
	return(stat);
}

int FLDdel_cell_set(field, cell_set)
    OMobj_id  field, cell_set;
{
	OMobj_id  elem_id;
	int stat;

	elem_id = OMfind_subobj (field, OMstr_to_name("cell_set"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);

	stat = OMdel_array_val (elem_id, cell_set);
	if (stat != 1)
		return(stat);
	return(OMdestroy_obj(cell_set, 0));
}

int FLDget_ncell_sets(field, ncell_sets)
    OMobj_id  field;
    int     *ncell_sets;
{
	return
		( OMget_name_int_val (field, OMstr_to_name("ncell_sets"), ncell_sets) );
}

int FLDset_ncell_sets (field, ncell_sets)
	OMobj_id  field;
	int ncell_sets;
{
	int   stat;
	OMobj_id   elem_id;

	stat = OMset_name_int_val (field, OMstr_to_name("ncell_sets"), ncell_sets);
	return(stat);
}

int FLDget_ncells(cell_set, ncells)
    OMobj_id  cell_set;
    int     *ncells;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("ncells"), ncells) );
}

int FLDset_ncells (cell_set, ncells)
	OMobj_id  cell_set;
	int ncells;
{
	int   stat;
	OMobj_id   elem_id;

	stat = OMset_name_int_val (cell_set, OMstr_to_name("ncells"), ncells);
	return(stat);
}

int FLDget_cell_ndim(cell_set, ndim)
    OMobj_id  cell_set;
    int     *ndim;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("cell_ndim"), ndim) );
}

int FLDget_cell_order(cell_set, order)
    OMobj_id  cell_set;
    int     *order;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("cell_order"), order) );
}

int FLDget_cell_type(cell_set, cell_type)
    OMobj_id  cell_set;
    int     *cell_type;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("cell_type"), cell_type) );
}

int FLDset_cell_type (cell_set, cell_type)
	OMobj_id  cell_set;
	int cell_type;
{
	int   stat;
	OMobj_id   elem_id;

	stat = OMset_name_int_val (cell_set, OMstr_to_name("cell_type"), cell_type);
	return(stat);
}

int FLDget_cell_set_nnodes(cell_set,  cell_nnodes)
    OMobj_id  cell_set;
    int  *cell_nnodes;
{
	return(OMget_name_int_val (cell_set, OMstr_to_name("cell_nnodes"), cell_nnodes));
	
}
int FLDget_cell_corner_nnodes(cell_set,  cell_nnodes)
    OMobj_id  cell_set;
    int  *cell_nnodes;
{
	if (OMget_name_int_val (cell_set, OMstr_to_name("cell_corner_nnodes"), 
				cell_nnodes) < 1)
		return(OMget_name_int_val (cell_set, OMstr_to_name("cell_nnodes"), cell_nnodes));
	else
		return(1);
}

int FLDget_cell_set_nfaces(cell_set,  cell_nfaces)
    OMobj_id  cell_set;
    int  *cell_nfaces;
{
	return(OMget_name_int_val (cell_set, OMstr_to_name("cell_nfaces"), cell_nfaces));
			
}

int FLDget_cell_faces(cell_set, nfaces, face_nnodes, face_corner_nnodes, node_list, size)
    OMobj_id  cell_set;
    int  **face_nnodes;
    int  **face_corner_nnodes;
    int  **node_list;
    int  *nfaces, *size;
{
	OMobj_id   elem_id;
	int     i, type, stat;
	int  mode;

	*node_list = (int *)NULL;
	*face_nnodes = (int *)NULL;
	*face_corner_nnodes = (int *)NULL;
	*size = 0;
	*nfaces = 0;
	type = OM_TYPE_INT;
	mode = OM_GET_ARRAY_RD;

	elem_id = OMfind_subobj (cell_set, OMstr_to_name("cell_face_nnodes"),OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)face_nnodes, nfaces, mode);
	if (stat < 1)
		return(stat);
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("cell_face_corner_nnodes"),OM_OBJ_RD);
	if (OMis_null_obj(elem_id))  {
		ARRincr_refcnt(*face_nnodes);
		*face_corner_nnodes = *face_nnodes;
	}
	else {
		stat = OMget_array_sz(elem_id, &type, (char **)face_corner_nnodes, nfaces, mode);
		if (stat < 1)
			return(stat);
	}
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("cell_face_connect"), OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)node_list, size, mode);
	if (stat < 1)
		return(stat);
	return(1);
}

int FLDget_node_connect(cell_set, node_conn_array, size, mode)
    OMobj_id  cell_set;
    int  **node_conn_array;
    int  *size;
    int  mode;
{
	
	OMobj_id   elem_id;
	int        type, stat;

	*node_conn_array = (int *)NULL;
	*size = 0;
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("node_connect_list"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)node_conn_array, size, mode);
	return (stat);
}

int FLDset_node_connect(cell_set, node_conn_array, size, mode)
    OMobj_id  cell_set;
    int size, mode;
    int *node_conn_array;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("node_connect_list"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)node_conn_array, size, mode);
	return(stat);
}


int FLDget_npolys(cell_set, npolys)
    OMobj_id  cell_set;
    int     *npolys;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("npolys"), npolys) );
}

int FLDset_npolys (cell_set, npolys)
	OMobj_id  cell_set;
	int npolys;
{
	int   stat;
	OMobj_id   elem_id;

	stat = OMset_name_int_val (cell_set, OMstr_to_name("npolys"), npolys);
	return(stat);
}

int FLDget_poly_flag(cell_set, poly_flag)
    OMobj_id  cell_set;
    int     *poly_flag;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("poly_flag"), poly_flag) );
}

int FLDset_poly_flag (cell_set, poly_flag)
	OMobj_id  cell_set;
	int poly_flag;
{
	int   stat;
	OMobj_id   elem_id;

	stat = OMset_name_int_val (cell_set, OMstr_to_name("poly_flag"), poly_flag);
	return(stat);
}

int FLDget_poly_connect(cell_set, poly_conn_array, size, mode)
    OMobj_id  cell_set;
    int  **poly_conn_array;
    int  *size;
    int  mode;
{
	
	OMobj_id   elem_id;
	int        type, stat;

	*poly_conn_array = (int *)NULL;
	*size = 0;
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("poly_connect_list"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)poly_conn_array, size, mode);
	return (stat);
}

int FLDset_poly_connect(cell_set, poly_conn_array, size, mode)
    OMobj_id  cell_set;
    int size, mode;
    int *poly_conn_array;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("poly_connect_list"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)poly_conn_array, size, mode);
	return(stat);
}


int FLDget_poly_nnodes(cell_set, poly_conn_array, size, mode)
    OMobj_id  cell_set;
    int  **poly_conn_array;
    int  *size;
    int  mode;
{
	
	OMobj_id   elem_id;
	int        type, stat;

	*poly_conn_array = (int *)NULL;
	*size = 0;
	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("poly_nnodes"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)poly_conn_array, size, mode);
	return (stat);
}

int FLDset_poly_nnodes(cell_set, poly_conn_array, size, mode)
    OMobj_id  cell_set;
    int size, mode;
    int *poly_conn_array;
{
	int   stat, type;
	OMobj_id   elem_id;

	type = OM_TYPE_INT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("poly_nnodes"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMset_array(elem_id, type, (char *)poly_conn_array, size, mode);
	return(stat);
}


int FLDget_cell_node_uvw(cell_set, uvw)
    OMobj_id  cell_set;
    float     **uvw;
{
	OMobj_id   elem_id;
	int        size, type, stat;
	
	*uvw = (float *)NULL;
	size = 0;
	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("node_uvw"),
				  OM_OBJ_RD);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)uvw, &size, OM_GET_ARRAY_RD);
	return (stat);
}

int FLDget_cell_nprops(cell_set, nprops)
    OMobj_id  cell_set;
    int     *nprops;
{
	return
		( OMget_name_int_val (cell_set, OMstr_to_name("nprops"), nprops) );
}

int FLDset_cell_nprops(cell_set, nprops)
    OMobj_id  cell_set;
    int     nprops;
{
	return
		( OMset_name_int_val (cell_set, OMstr_to_name("nprops"), nprops) );
}

int FLDget_cell_props(cell_set, props, size, mode)
    OMobj_id  cell_set;
    float  **props;
    int  *size;
    int  mode;
{
	
	OMobj_id   elem_id;
	int        type, stat;

	*props = (float *)NULL;
	*size = 0;
	type = OM_TYPE_FLOAT;
	elem_id = OMfind_subobj (cell_set, OMstr_to_name("props"),
				   OM_OBJ_RW);
	if (OMis_null_obj(elem_id)) 
		return(0);
	stat = OMget_array_sz(elem_id, &type, (char **)props, size, mode);
	return (stat);
}

int FLDcheck_cell_connect(cell_set, nnodes, bad_index, bad_value)
    OMobj_id  cell_set;
    int  nnodes;
    int  *bad_index, *bad_value;
{
	int stat, poly_flag, size, *node_conn;

	if (FLDget_poly_flag(cell_set, &poly_flag) != 1) {
		poly_flag = 0;
	}
	
	if (poly_flag == 0) {
		stat = FLDget_node_connect(cell_set, &node_conn, &size, OM_GET_ARRAY_RD);
		if (stat != 1)
			return(0);
	}
	else {
		stat = FLDget_poly_connect(cell_set, &node_conn, &size, OM_GET_ARRAY_RD);
		if (stat != 1) 
			return(0);
	}
	stat = UTILcheck_list(node_conn, size, 0, nnodes-1, bad_index, bad_value);
	ARRfree(node_conn);
	if (stat != 1) 
		return(-1);
	return(1);
}

int FLDcheck_mesh_connect(field, bad_index, bad_value)
    OMobj_id  field;
    int  *bad_index, *bad_value;
{
	int stat, set, nsets, nnodes, ndim;
	OMobj_id  cell_set;

	*bad_index = 0;
	*bad_value = 0;
	stat = FLDget_ndim(field, &ndim);
	if (stat >= 0) /** dont check Struct meshes **/
		return(1);
	stat = FLDget_ncell_sets(field, &nsets);
	if (stat != 1)
		return(0);
	stat = FLDget_nnodes(field, &nnodes);
	if (stat != 1)
		return(0);
	for (set=0; set<nsets; set++) {
		stat = FLDget_cell_set(field, set, &cell_set);
		if (stat != 1)
			return(0);
		stat = FLDcheck_cell_connect(cell_set, nnodes, bad_index, bad_value);
		if (stat < 0)
			return(stat);
	}
	return(1);
}

/********************************/
/*  Time dependent data API     */
/********************************/


/*----------------------------------------------*/
/*  returns number of time steps in the field   */
/*----------------------------------------------*/
int FLDget_time_nsteps(OMobj_id field, int *nsteps)
{
	return (OMget_name_int_val (field, OMstr_to_name("nsteps"), nsteps) );
}

/*----------------------------------------------*/
/*  sets number of time steps in the field      */
/*----------------------------------------------*/
int FLDset_time_nsteps (OMobj_id field, int nsteps)
{
	return (OMset_name_int_val (field, OMstr_to_name("nsteps"), nsteps));
}

/*----------------------------------------------*/
/*  returns array of time values in the field   */
/*----------------------------------------------*/
int FLDget_time_values(OMobj_id field, double **time, int *size, int mode)
{
	OMobj_id  id;
	int    type, stat, nnodes, nspace;
	
	type = OM_TYPE_DOUBLE;
	*time = (double *)NULL;
	*size = 0;
	id = OMfind_subobj (field, OMstr_to_name("time"), 
			    OM_OBJ_RW);
	if (OMis_null_obj(id)) 
		return(0);
	stat=OMget_array_sz(id, &type, (char **)time, size, mode);
	return(stat);
}

/*----------------------------------------------*/
/*  sets array of time values in the field      */
/*----------------------------------------------*/
int FLDset_time_values(OMobj_id field, double *time, int size, int mode)
{
	int   stat, type;
	OMobj_id   id;

	type = OM_TYPE_DOUBLE;
	id = OMfind_subobj (field, OMstr_to_name("time"),
				   OM_OBJ_RW);
	if (OMis_null_obj(id)) 
		return(0);
	stat = OMset_array(id, type, (char *)time, size, mode);
	return(stat);
}

/*----------------------------------------------*/
/*  returns id of one coordinate time step      */
/*----------------------------------------------*/
int FLDget_time_coordinates(OMobj_id field, int step, OMobj_id *coordinates)
{
	OMobj_id   id;
	
	id = OMfind_subobj (field, OMstr_to_name("time_coordinates"), OM_OBJ_RW);
	if (OMis_null_obj(id)) 
		return(0);
	if (OMget_array_val (id, step, coordinates, OM_OBJ_RW) <= 0)
		return(0);
	return(1);
}

/*----------------------------------------------*/
/*  returns id of one node_data time step       */
/*----------------------------------------------*/
int FLDget_time_node_data(OMobj_id field, int step, OMobj_id *node_data)
{
	OMobj_id   id;
	
	id = OMfind_subobj (field, OMstr_to_name("time_node_data"), OM_OBJ_RW);
	if (OMis_null_obj(id)) 
		return(0);
	if (OMget_array_val (id, step, node_data, OM_OBJ_RW) <= 0)
		return(0);
	return(1);
}


/*----------------------------------------------*/
/*  returns id of one cell_data time step       */
/*----------------------------------------------*/
int FLDget_time_cell_data(OMobj_id cell_set, int step, OMobj_id *cell_data)
{
	OMobj_id   id;
	
	id = OMfind_subobj (cell_set, OMstr_to_name("time_cell_data"), OM_OBJ_RW);
	if (OMis_null_obj(id)) 
		return(0);
	if (OMget_array_val (id, step, cell_data, OM_OBJ_RW) <= 0)
		return(0);
	return(1);
}




